#include "./controller/widget/widget_node_view.h"

#include "./controller/controller_gui.h"
#include "./core/factory.h"
#include "./core/scene.h"

namespace Splash
{

/*************/
GuiNodeView::GuiNodeView(Scene* scene, const std::string& name)
    : GuiWidget(scene, name)
{
}

/*************/
void GuiNodeView::initializeNodePositions(const std::vector<std::string>& objectNames)
{
    // Get the link level for each object, corresponding to the level of descendants
    std::map<std::string, uint32_t> descendants;
    {
        const auto links = getObjectLinks();
        const auto reversedLinks = getObjectReversedLinks();

        // Get the forward level
        for (const auto& objectName : objectNames)
        {
            std::vector<std::string> crossedObject;
            std::vector<std::string> currentLinks = links.at(objectName);

            uint32_t level = 0;
            while (!currentLinks.empty())
            {
                ++level;

                // Generate the list of objects at the next level
                std::vector<std::string> nextLinks;
                for (const auto& link : currentLinks)
                {
                    // If the link is not to an object mentioned in objectNames, do not add its own links
                    if (std::find(objectNames.cbegin(), objectNames.cend(), link) == objectNames.cend())
                        continue;

                    // If the link is to an object already visited, do not add it either
                    if (std::find(crossedObject.cbegin(), crossedObject.cend(), link) != crossedObject.cend())
                        continue;

                    crossedObject.push_back(link);
                    std::transform(links.at(link).cbegin(), links.at(link).cend(), std::back_inserter(nextLinks), [](const auto& name) { return name; });
                }
                std::swap(currentLinks, nextLinks);
            }

            descendants[objectName] = level;

            // We leave non-linked objects at the root level, by themselves
            if (!links.at(objectName).empty() || !reversedLinks.at(objectName).empty())
                descendants[objectName] += 1;
        }
    }

    // Evaluate the number of objects per level
    std::map<int, int> countPerLevel;
    int maxCountPerLevel = 0;
    for (const auto& object : descendants)
    {
        auto count = countPerLevel.find(object.second);
        if (count == countPerLevel.end())
        {
            countPerLevel[object.second] = 0;
            count = countPerLevel.find(object.second);
        }
        count->second++;
        maxCountPerLevel = std::max(maxCountPerLevel, count->second);
    }

    // Compute nodes position
    static ImVec2 nodeMargin{32, 32};
    std::map<int, float> positionPerLevel;
    for (const auto& objectName : objectNames)
    {
        auto level = descendants.find(objectName);
        assert(level != descendants.end());

        auto position = positionPerLevel.find(level->second);
        if (position == positionPerLevel.end())
        {
            positionPerLevel[level->second] = 0;
            position = positionPerLevel.find(level->second);
        }
        else
        {
            position->second += 1;
        }

        auto diffToMaxCount = static_cast<float>(maxCountPerLevel - countPerLevel[level->second]);

        auto xPosition = static_cast<float>(level->second) * (_nodeSize[0] + nodeMargin[0]);
        auto yPosition = (static_cast<float>(position->second) + diffToMaxCount / 2.f) * (_nodeSize[1] + nodeMargin[1]);
        _nodePositions[objectName] = std::vector<float>({xPosition, yPosition});
    }
}

/*************/
void GuiNodeView::render()
{
    const ImGuiIO& io = ImGui::GetIO();

    auto objectLinks = getObjectLinks();
    auto objectNames = getObjectList();
    auto objectAliases = getObjectAliases();

    if (!_viewNonSavableObjects)
    {
        objectNames.erase(std::remove_if(objectNames.begin(),
                              objectNames.end(),
                              [this](const auto& name) -> bool {
                                  Values savable = getObjectAttribute(name, "savable");
                                  if (savable.empty())
                                      return true;
                                  return !savable[0].as<bool>();
                              }),
            objectNames.end());
    }
    _clickedNode.clear();

    // A child window to be able to insert the menu
    ImGui::BeginChild("GraphArea", ImVec2(0, 0), ImGuiChildFlags_None, ImGuiWindowFlags_NoScrollbar | ImGuiWindowFlags_NoScrollWithMouse | ImGuiWindowFlags_MenuBar);
    if (ImGui::BeginMenuBar())
    {
        if (ImGui::BeginMenu("View"))
        {
            if (ImGui::MenuItem("Reorder graph"))
                initializeNodePositions(objectNames);
            if (ImGui::IsItemHovered())
                ImGui::SetTooltip("Organize nodes by their links (R)");

            if (ImGui::MenuItem("Show/hide non-savable objects"))
                _viewNonSavableObjects = !_viewNonSavableObjects;
            if (ImGui::IsItemHovered())
                ImGui::SetTooltip("Show/hide objects generated at runtime (H)");

            ImGui::EndMenu();
        }
        if (ImGui::BeginMenu("Select"))
        {
            auto addMenuForCategory = [&, this](const std::string& categoryName, GraphObject::Category category) {
                if (ImGui::BeginMenu(categoryName.c_str()))
                {
                    const auto types = getTypesFromCategory(category);
                    for (const auto& name : objectNames)
                    {
                        const auto object = getObjectPtr(name);
                        if (object == nullptr)
                            continue;
                        const auto objectType = object->getType();
                        if (std::find(types.cbegin(), types.cend(), objectType) == types.cend())
                            continue;

                        const auto& alias = objectAliases[name];
                        if (ImGui::MenuItem(alias.c_str()))
                        {
                            _sourceNode = name;
                            _clickedNode = _sourceNode;

                            // Move the focus to the selected object
                            auto nodePos = _nodePositions.find(_sourceNode);
                            assert(nodePos != _nodePositions.end());
                            _viewShift[0] = -nodePos->second[0] + _graphSize.x / 2.f - _nodeSize[0] / 2.f;
                            _viewShift[1] = -nodePos->second[1] + _graphSize.y / 2.f - _nodeSize[1] / 2.f;
                        }
                    }

                    ImGui::EndMenu();
                }
            };

            addMenuForCategory("Meshes", GraphObject::Category::MESH);
            addMenuForCategory("Images", GraphObject::Category::IMAGE);
            addMenuForCategory("Filters", GraphObject::Category::FILTER);
            addMenuForCategory("Sinks", GraphObject::Category::SINK);
            addMenuForCategory("Others", GraphObject::Category::MISC);

            ImGui::EndMenu();
        }
        if (ImGui::IsItemHovered())
            ImGui::SetTooltip("Select an object in the graph");
        if (ImGui::BeginMenu("Add"))
        {
            renderAddObjectMenu();
            ImGui::EndMenu();
        }
        if (ImGui::IsItemHovered())
            ImGui::SetTooltip("Add an object to the graph");
        if (ImGui::BeginMenu("Object"))
        {
            if (ImGui::MenuItem("Delete"))
                setWorldAttribute("deleteObject", {_sourceNode});
            if (ImGui::IsItemHovered())
                ImGui::SetTooltip("Delete, no question asked (delete/backspace when over the nodes)");
            ImGui::EndMenu();
        }
        ImGui::EndMenuBar();
    }

    // Begin a subwindow to enclose nodes
    ImGui::BeginChild("NodeView", ImVec2(0.f, 0.f), ImGuiChildFlags_Borders, ImGuiWindowFlags_NoScrollbar | ImGuiWindowFlags_NoScrollWithMouse);

    if (_firstRender)
    {
        // First render for a node: organize it given its links
        initializeNodePositions(objectNames);
        _firstRender = false;
    }
    else
    {
        // For subsequent renders: compute the position for new nodes only
        for (const auto& objectName : objectNames)
        {
            if (_nodePositions.find(objectName) != _nodePositions.end())
                continue;
            _nodePositions[objectName] = std::vector<float>({-_viewShift[0] + _graphSize.x / 2.f - _nodeSize[0] / 2.f, -_viewShift[1] + _graphSize.y / 2.f - _nodeSize[1] / 2.f});
        }
    }

    // Apply view shift
    auto canvasPos = ImGui::GetCursorScreenPos();
    canvasPos.x += _viewShift[0];
    canvasPos.y += _viewShift[1];

    // Draw objects according to their link level
    for (const auto& objectName : objectNames)
        renderNode(objectName);

    // Draw lines
    const auto& style = ImGui::GetStyle();
    auto drawList = ImGui::GetWindowDrawList();
    for (const auto& object : objectLinks)
    {
        const auto& name = object.first;
        const auto& links = object.second;

        // Check that the source is displayed
        if (std::find(objectNames.begin(), objectNames.end(), name) == objectNames.end())
            continue;

        if (links.empty())
            continue;

        const auto& currentNodePos = _nodePositions[name];
        const auto p1 = ImVec2(currentNodePos[0] + canvasPos.x - style.WindowPadding[0], currentNodePos[1] + canvasPos.y - style.WindowPadding[1] + _nodeSize[1] / 2.0);

        for (const auto& target : links)
        {
            // Check that the source is displayed
            if (std::find(objectNames.begin(), objectNames.end(), target) == objectNames.end())
                continue;

            const auto targetIt = _nodePositions.find(target);
            if (targetIt == _nodePositions.end())
                continue;

            const auto& targetPos = targetIt->second;
            const auto p4 = ImVec2(targetPos[0] + canvasPos.x + _nodeSize[0] - style.WindowPadding[0], targetPos[1] + canvasPos.y - style.WindowPadding[1] + _nodeSize[1] / 2.0);
            const auto p2 = ImVec2(p1.x * 0.66f + p4.x * 0.33f, p1.y);
            const auto p3 = ImVec2(p1.x * 0.33f + p4.x * 0.66f, p4.y);

            drawList->AddBezierCubic(p1, p2, p3, p4, 0xBB0088FF, 2.f, 16);
        }
    }

    // end of the subwindow
    ImGui::EndChild(); // NodeView
    _graphSize = ImGui::GetItemRectSize();

    // Interactions
    if (ImGui::IsItemHovered())
        _isHovered = true;
    else
        _isHovered = false;

    // This handles the mouse capture even when the mouse goes outside the node view widget
    static bool viewCaptured = false;
    if (io.MouseDownDuration[2] > 0.0)
    {
        if (_isHovered)
            viewCaptured = true;
    }
    else if (viewCaptured)
    {
        viewCaptured = false;
    }

    if (viewCaptured)
    {
        _viewShift[0] += io.MouseDelta.x;
        _viewShift[1] += io.MouseDelta.y;
    }

    if (_isHovered)
    {
        // The add menu is also rendered when right clicking in the node view
        if (io.MouseClicked[1])
        {
            const auto clickPosition = io.MouseClickedPos[1];
            ImGui::SetCursorScreenPos(clickPosition);
            ImGui::OpenPopup("addObjectPopup");
        }

        // A few keyboard shortcuts
        if (ImGui::IsKeyPressed(ImGuiKey_Delete, false) || ImGui::IsKeyPressed(ImGuiKey_Backspace, false))
        {
            setWorldAttribute("deleteObject", {_sourceNode});
            _sourceNode.clear();
        }
        else if (ImGui::IsKeyPressed(ImGuiKey_R, false))
        {
            initializeNodePositions(objectNames);
        }
        else if (ImGui::IsKeyPressed(ImGuiKey_H, false))
        {
            _viewNonSavableObjects = !_viewNonSavableObjects;
        }
    }

    // Drawing the right-click menu
    if (ImGui::BeginPopup("addObjectPopup"))
    {
        renderAddObjectMenu();
        ImGui::EndPopup();
    }

    ImGui::EndChild(); // GraphArea

    auto currentPos = ImGui::GetCursorScreenPos();
    currentPos.x += 8;
    currentPos.y -= 32;
    ImGui::SetCursorScreenPos(currentPos);

    ImGui::PushFont(Gui::guiFonts[Gui::FontType::Bold]);
    ImGui::Text("Click: select");
    ImGui::PopFont();
    ImGui::SameLine();
    ImGui::Text("- Shift + click: link -");

    ImGui::SameLine();
    ImGui::PushFont(Gui::guiFonts[Gui::FontType::Bold]);
    ImGui::Text("Ctrl + click: unlink");
    ImGui::PopFont();
    ImGui::SameLine();
    ImGui::Text("- Middle click or Alt + right click: pan -");

    ImGui::SameLine();
    ImGui::PushFont(Gui::guiFonts[Gui::FontType::Bold]);
    ImGui::Text("Right click: add object");
    ImGui::PopFont();
}

/*************/
void GuiNodeView::renderAddObjectMenu()
{
    auto drawTypesMenu = [&](const auto& types) {
        for (const auto& type : types)
        {
            const auto desc = type + " : " + getShortDescription(type);
            if (ImGui::MenuItem(desc.c_str(), nullptr))
                setWorldAttribute("addObject", {type});
            if (ImGui::IsItemHovered())
            {
                const auto fullDesc = getDescription(type);
                ImGui::SetTooltip("%s", fullDesc.c_str());
            }
        }
    };

    if (ImGui::BeginMenu("Mesh"))
    {
        const auto types = getTypesFromCategory(GraphObject::Category::MESH);
        drawTypesMenu(types);
        ImGui::EndMenu();
    }
    if (ImGui::BeginMenu("Image"))
    {
        const auto types = getTypesFromCategory(GraphObject::Category::IMAGE);
        drawTypesMenu(types);
        ImGui::EndMenu();
    }
    if (ImGui::BeginMenu("Filter"))
    {
        const auto types = getTypesFromCategory(GraphObject::Category::FILTER);
        drawTypesMenu(types);
        ImGui::EndMenu();
    }
    if (ImGui::BeginMenu("Sink"))
    {
        const auto types = getTypesFromCategory(GraphObject::Category::SINK);
        drawTypesMenu(types);
        ImGui::EndMenu();
    }
    if (ImGui::BeginMenu("Other"))
    {
        const auto types = getTypesFromCategory(GraphObject::Category::MISC);
        drawTypesMenu(types);
        ImGui::EndMenu();
    }
}

/*************/
void GuiNodeView::renderNode(const std::string& name)
{
    const auto& io = ImGui::GetIO();
    auto pos = _nodePositions.find(name);
    // If the node does not have a drawing position, do not draw it
    if (pos == _nodePositions.end())
        return;

    auto& nodePos = pos->second;
    ImGui::SetCursorPos(ImVec2(nodePos[0] + _viewShift[0], nodePos[1] + _viewShift[1]));

    // Beginning of node rendering
    bool coloredSelectedNode = false;
    if (name == _sourceNode)
    {
        ImGui::PushStyleColor(ImGuiCol_Button, ImGui::GetStyle().Colors[ImGuiCol_ButtonActive]);
        coloredSelectedNode = true;
    }

    ImGui::BeginChild(std::string("node_" + name).c_str(), ImVec2(_nodeSize[0], _nodeSize[1]), ImGuiChildFlags_None);
    ImGui::Button(getObjectAlias(name).c_str(), ImVec2(_nodeSize[0], _nodeSize[1]));
    if (ImGui::IsItemHovered())
    {
        if (io.MouseClicked[0])
        {
            // Object linking
            if (io.KeyShift)
            {
                if (dynamic_cast<Scene*>(_root))
                    setWorldAttribute("link", {_sourceNode, name});
            }
            else if (io.KeyCtrl)
            {
                if (dynamic_cast<Scene*>(_root))
                    setWorldAttribute("unlink", {_sourceNode, name});
            }
            else
            {
                _clickedNode = name;
                _sourceNode = name;
            }
        }
        else if (io.MouseDownDuration[0] > 0.0)
        {
            // Dragging
            _capturedNode = name;
        }
    }

    if (io.MouseReleased[0])
        _capturedNode.clear();

    if (_capturedNode == name)
    {
        const float dx = io.MouseDelta.x;
        const float dy = io.MouseDelta.y;
        nodePos[0] += dx;
        nodePos[1] += dy;
    }

    ImGui::EndChild();

    if (coloredSelectedNode)
        ImGui::PopStyleColor(1);
}

/*************/
int GuiNodeView::updateWindowFlags()
{
    ImGuiWindowFlags flags = 0;
    if (_isHovered)
    {
        flags |= ImGuiWindowFlags_NoMove;
    }
    return flags;
}

} // namespace Splash
