#include "./controller/widget/widget_stats.h"

#include <imgui.h>

#include "./utils/timer.h"

namespace Splash
{

/*************/
void GuiStats::render()
{
    static ImGuiTableFlags flags = ImGuiTableFlags_BordersV | ImGuiTableFlags_BordersOuterH | ImGuiTableFlags_Resizable | ImGuiTableFlags_RowBg | ImGuiTableFlags_NoBordersInBody;
    static ImGuiTreeNodeFlags treeNodeFlags = ImGuiTreeNodeFlags_SpanAllColumns;
    if (ImGui::BeginTable("Statistics", 3, flags))
    {
        const float textBaseWidth = ImGui::CalcTextSize("A").x;
        ImGui::TableSetupColumn("Counters", ImGuiTableColumnFlags_NoHide);
        ImGui::TableSetupColumn("Value", ImGuiTableColumnFlags_WidthFixed, textBaseWidth * 12.f);
        ImGui::TableSetupColumn("Unit", ImGuiTableColumnFlags_WidthFixed, textBaseWidth * 4.f);
        ImGui::TableHeadersRow();

        // The timings are organized in a tree structure
        struct Level
        {
            std::string name = "";
            float value = 0.f;
            std::string unit = "";
            std::vector<Level> sublevels = {};
        };

        std::function<void(const Level& level)> displayLevel;
        displayLevel = [&](const Level& level) {
            ImGui::TableNextRow();
            ImGui::TableNextColumn();
            if (level.sublevels.empty())
            {
                ImGui::TreeNodeEx(level.name.c_str(), treeNodeFlags | ImGuiTreeNodeFlags_Leaf | ImGuiTreeNodeFlags_Bullet | ImGuiTreeNodeFlags_NoTreePushOnOpen);
                ImGui::TableNextColumn();
                ImGui::Text("%f", level.value);
                ImGui::TableNextColumn();
                ImGui::TextUnformatted(level.unit.c_str());
            }
            else
            {
                const bool isOpen = ImGui::TreeNodeEx(level.name.c_str(), treeNodeFlags);
                ImGui::TableNextColumn();
                ImGui::Text("%f", level.value);
                ImGui::TableNextColumn();
                ImGui::TextUnformatted(level.unit.c_str());
                if (isOpen)
                {
                    for (const auto& sublevel : level.sublevels)
                        displayLevel(sublevel);
                    ImGui::TreePop();
                }
            }
        };

        // Prepare the data to be displayed
        static std::unordered_map<std::string, float> stats;

        const auto tree = _root->getTree();
        auto runningAverage = [](float a, float b) { return a * 0.9 + 0.0001 * b; }; // 0.1 * 0.001, to include conversion from ms to sec
        auto getLeafValue = [&](const std::string& path) {
            Value value;
            if (!tree->getValueForLeafAt(path, value))
                return 0.f;
            return value[0].as<float>();
        };

        // Process the world first
        {
            assert(tree->hasBranchAt("/world/durations"));

            stats["world_loop_world"] = runningAverage(stats["world_loop_world"], getLeafValue("/world/durations/loop_world"));
            stats["world_loop_world_fps"] = 1e3 / std::max(1.f, stats["world_loop_world"]);
            stats["world_upload"] = runningAverage(stats["world_upload"], getLeafValue("/world/durations/upload"));

            Level world = {
                .name = "World",
                .value = stats["world_loop_world_fps"],
                .unit = "fps",
            };
            world.sublevels.push_back({.name = "Duration per frame", .value = stats["world_loop_world"], .unit = "ms"});
            world.sublevels.push_back({.name = "Buffers upload", .value = stats["world_upload"], .unit = "ms"});

            ImGui::SetNextItemOpen(true);
            displayLevel(world);
        }

        // Process the scenes
        for (const auto& branchName : tree->getBranchList())
        {
            if (branchName == "world")
                continue;
            const auto durationPath = "/" + branchName + "/durations";

            stats[branchName + "_loop_scene"] = runningAverage(stats[branchName + "_loop_scene"], getLeafValue(durationPath + "/loop_scene"));
            stats[branchName + "_loop_scene_fps"] = 1e3 / std::max(1.f, stats[branchName + "_loop_scene"]);
            stats[branchName + "_textureUpload"] = runningAverage(stats[branchName + "_textureUpload"], getLeafValue(durationPath + "/textureUpload"));
            stats[branchName + "_blender"] = runningAverage(stats[branchName + "_blender"], getLeafValue(durationPath + "/blender"));
            stats[branchName + "_filter"] = runningAverage(stats[branchName + "_filter"], getLeafValue(durationPath + "/filter"));
            stats[branchName + "_camera"] = runningAverage(stats[branchName + "_camera"], getLeafValue(durationPath + "/camera"));
            stats[branchName + "_warp"] = runningAverage(stats[branchName + "_warp"], getLeafValue(durationPath + "/warp"));
            stats[branchName + "_renderingContext"] = runningAverage(stats[branchName + "_renderingContext"], getLeafValue(durationPath + "/window"));
            stats[branchName + "_swap"] = runningAverage(stats[branchName + "_swap"], getLeafValue(durationPath + "/swap"));
            stats[branchName + "_gl_time_per_frame"] =
                runningAverage(stats[branchName + "_gl_time_per_frame"], getLeafValue(durationPath + "/" + Constants::GL_TIMING_PREFIX + Constants::GL_TIMING_TIME_PER_FRAME));
            stats[branchName + "_gui"] = runningAverage(stats[branchName + "_gui"], getLeafValue(durationPath + "/gui"));

            Level scene = {
                .name = branchName,
                .value = stats[branchName + "_loop_scene_fps"],
                .unit = "fps",
            };

            Level gpu = {
                .name = "Time per frame on the GPU (graphic card)",
                .value = stats[branchName + "_gl_time_per_frame"],
                .unit = "ms",
            };

            Level cpu = {
                .name = "CPU (processor)",
                .value = stats[branchName + "_loop_scene_fps"],
                .unit = "fps",
            };

            cpu.sublevels.push_back({.name = "Time per frame", .value = stats[branchName + "_loop_scene"], .unit = "ms"});
            cpu.sublevels.push_back({.name = "Texture upload", .value = stats[branchName + "_textureUpload"], .unit = "ms"});
            cpu.sublevels.push_back({.name = "Projection blending", .value = stats[branchName + "_blender"], .unit = "ms"});
            cpu.sublevels.push_back({.name = "Filters", .value = stats[branchName + "_filter"], .unit = "ms"});
            cpu.sublevels.push_back({.name = "Cameras", .value = stats[branchName + "_camera"], .unit = "ms"});
            cpu.sublevels.push_back({.name = "Warps", .value = stats[branchName + "_warp"], .unit = "ms"});
            cpu.sublevels.push_back({.name = "Windows", .value = stats[branchName + "_renderingContext"], .unit = "ms"});
            cpu.sublevels.push_back({.name = "Buffer swapping", .value = stats[branchName + "_swap"], .unit = "ms"});
            cpu.sublevels.push_back({.name = "User interface", .value = stats[branchName + "_gui"], .unit = "ms"});

            scene.sublevels.push_back(gpu);
            scene.sublevels.push_back(cpu);

            ImGui::SetNextItemOpen(true);
            displayLevel(scene);
        }

        ImGui::EndTable();
    }
}

} // namespace Splash
