#include "./controller/widget/widget_textures.h"

#include <algorithm>

#include <imgui.h>

#include "./graphics/filter.h"

namespace Splash
{
/*************/
std::function<std::vector<std::shared_ptr<Texture>>()> GuiTextures::getTextureLister(GraphObject::Category category)
{
    if (category == GraphObject::Category::FILTER)
    {
        auto getTextures = [&, this]() -> std::vector<std::shared_ptr<Texture>> {
            const auto objectsAsFilters = getObjectsOfBaseType<Filter>();
            std::vector<std::shared_ptr<Texture>> objects;
            for (auto& obj : objectsAsFilters)
            {
                auto objAsTexture = std::dynamic_pointer_cast<Texture>(obj);
                if (objAsTexture)
                    objects.push_back(objAsTexture);
            }
            return objects;
        };
        return getTextures;
    }
    else if (category == GraphObject::Category::TEXTURE)
    {
        return [&, this]() -> std::vector<std::shared_ptr<Texture>> {
            return getObjectsOfBaseType<Texture>();
        };
    }

    assert(false);
    return []() -> std::vector<std::shared_ptr<Texture>> { return {}; };
}

/*************/
void GuiTextures::render()
{
    ImGui::BeginChild("##textureArea", ImVec2(0, 0), ImGuiChildFlags_None, ImGuiWindowFlags_NoScrollbar | ImGuiWindowFlags_NoScrollWithMouse | ImGuiWindowFlags_MenuBar);
    if (ImGui::BeginMenuBar())
    {
        if (ImGui::BeginMenu("View"))
        {
            ImGui::Checkbox("Hide non-savable textures", &_hideNonSavable);
            if (ImGui::IsItemHovered())
                ImGui::SetTooltip("Hide the textures generated at runtime, whose attributes will not be saved");
            ImGui::EndMenu();
        }

        if (ImGui::BeginMenu("Filter by type"))
        {
            if (ImGui::MenuItem("Filters"))
                _selectedCategory = GraphObject::Category::FILTER;
            if (ImGui::IsItemHovered())
                ImGui::SetTooltip("View all filters");

            if (ImGui::MenuItem("All textures"))
                _selectedCategory = GraphObject::Category::TEXTURE;
            if (ImGui::IsItemHovered())
                ImGui::SetTooltip("View all textures");

            ImGui::EndMenu();
        }
        // Menu items should be added there, if needed
        ImGui::EndMenuBar();
    }

    const auto textureLister = getTextureLister(_selectedCategory);
    auto textureList = textureLister();

    if (_hideNonSavable)
        textureList.erase(std::remove_if(textureList.begin(), textureList.end(), [](const auto& texture) { return !texture->getSavable(); }), textureList.end());

    if (textureList.size() != 0 && _selectedObjectName.empty())
        _selectedObjectName = textureList[0]->getName();

    ImVec2 availableSize = ImGui::GetContentRegionAvail();
    ImGui::BeginChild("##texture", ImVec2(availableSize.x * 0.25, availableSize.y), ImGuiChildFlags_Borders);
    ImGui::Text("Texture list");

    const auto leftMargin = static_cast<int>(ImGui::GetCursorScreenPos().x - ImGui::GetWindowPos().x);
    availableSize = ImGui::GetContentRegionAvail();
    for (const auto& texture : textureList)
    {
        auto spec = texture->getSpec();

        int w = availableSize.x - 2 * leftMargin;
        int h = w * spec.height / spec.width;

        if (ImGui::ImageButton(texture->getName().c_str(), (ImTextureID)(intptr_t)texture->getTexId(), ImVec2(w, h), ImVec2(0, 0), ImVec2(1, 1)))
        {
            _selectedObjectName = texture->getName();
            _zoomRatio = 1.f;
        }
        if (ImGui::IsItemHovered())
            ImGui::SetTooltip("%s", texture->getAlias().c_str());
    }
    ImGui::EndChild(); // ##texture

    auto texture = std::dynamic_pointer_cast<Texture>(getObjectPtr(_selectedObjectName));
    if (texture)
    {
        ImGui::SameLine();
        ImGui::BeginChild("##textureInfo", ImVec2(0, 0), ImGuiChildFlags_Borders, ImGuiWindowFlags_HorizontalScrollbar);

        const auto spec = texture->getSpec();
        const auto ratio = static_cast<float>(spec.height) / static_cast<float>(spec.width);
        const int width = (ImGui::GetWindowWidth() / _zoomRatio) - ImGui::GetStyle().ScrollbarSize * 2.f;
        const int height = width * ratio;

        ImGui::Image((ImTextureID)(intptr_t)(texture->getTexId()), ImVec2(width, height));

        const auto& io = ImGui::GetIO();
        if (ImGui::IsItemHovered())
        {
            if (io.MouseWheel != 0.f)
                _zoomRatio = std::min(1.f, _zoomRatio * (1.f + io.MouseWheel * 0.04f));
        }

        // Move the texture around
        // viewCaptured is used to continue moving even if we
        // stop hovering the texture while doing it
        static bool viewCaptured = false;
        if (io.MouseDownDuration[2] > 0.f)
        {
            if (ImGui::IsItemHovered())
                viewCaptured = true;
        }
        else if (viewCaptured)
        {
            viewCaptured = false;
        }

        if (viewCaptured)
        {
            if (io.MouseDown[2])
            {
                auto scrollX = ImGui::GetScrollX();
                auto scrollY = ImGui::GetScrollY();
                const auto scrollMaxX = ImGui::GetScrollMaxX();
                const auto scrollMaxY = ImGui::GetScrollMaxY();

                float dx = io.MouseDelta.x;
                float dy = io.MouseDelta.y;
                scrollX = std::max(0.f, std::min(scrollMaxX, scrollX - dx));
                scrollY = std::max(0.f, std::min(scrollMaxY, scrollY - dy));

                ImGui::SetScrollX(scrollX);
                ImGui::SetScrollY(scrollY);
            }
        }

        ImGui::EndChild(); // ##textureInfo
    }

    ImGui::EndChild(); // ##textureArea
}

} // namespace Splash
